/*
 *  R : A Computer Language for Statistical Data Analysis
 *  Copyright (C) 1995, 1996  Robert Gentleman and Ross Ihaka
 *  Copyright (C) 1997--2003  Robert Gentleman, Ross Ihaka and the
 *			      R Development Core Team
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#define R_XGD_DEVICE 1
#include "devXGD.h"

/********************************************************/
/* If there are resources that are shared by all devices*/
/* of this type, you may wish to make them globals	*/
/* rather than including them in the device-specific	*/
/* parameters structure (especially if they are large !)*/
/********************************************************/

/* XGD Driver Specific parameters
 * with only one copy for all xGD devices */


/*  XGD Device Driver Arguments	:	*/
/*	1) display name			*/
/*	2) width (pixels)		*/
/*	3) height (pixels)		*/
/*	4) host to connect to		*/
/*	5) tcp port to connect to	*/

Rboolean newXGD_Open(NewDevDesc *dd, newXGDDesc *xd,
		     char *dsp, double w, double h, char *host, int port);
void setupXGDfunctions(NewDevDesc *dd);
int Rf_setNewXGDDeviceData(NewDevDesc *dd, double gamma_fac, newXGDDesc *xd);

Rboolean newXGDDeviceDriver(NewDevDesc *dd,
			    char *disp_name,
			    double width,
			    double height,
			    char *host,
			    int port)
{
  newXGDDesc *xd;
  char *fn;

#ifdef XGD_DEBUG
  printf("TD: newXGDDeviceDriver(\"%s\", %f, %f, \"%s\", %d)\n",disp_name, width,height,host,port);
#endif
  
  xd = Rf_allocNewXGDDeviceDesc(10);
  if (!newXGD_Open((NewDevDesc*)(dd), xd, disp_name, width, height,
		   host, port)) {
    free(xd);
    return FALSE;
  }
  
  Rf_setNewXGDDeviceData((NewDevDesc*)(dd), 0.6, xd);
  
  return TRUE;
}

/**
  This fills the general device structure (dd) with the XGD-specific
  methods/functions. It also specifies the current values of the
  dimensions of the device, and establishes the fonts, line styles, etc.
 */
int
Rf_setNewXGDDeviceData(NewDevDesc *dd, double gamma_fac, newXGDDesc *xd)
{
    /*	Set up Data Structures. */
    setupXGDfunctions(dd);

    /* Set required graphics parameters. */

    /* Window Dimensions in Pixels */
    /* Initialise the clipping rect too */

    dd->left = dd->clipLeft = 0;			/* left */
    dd->right = dd->clipRight = xd->windowWidth;	/* right */
    dd->bottom = dd->clipBottom = xd->windowHeight;	/* bottom */
    dd->top = dd->clipTop = 0;			/* top */

    /* Nominal Character Sizes in Pixels */

    dd->cra[0] = 8;
    dd->cra[1] = 8;

    /* Character Addressing Offsets */
    /* These are used to plot a single plotting character */
    /* so that it is exactly over the plotting point */

    dd->xCharOffset = 0.4900;
    dd->yCharOffset = 0.3333;
    dd->yLineBias = 0.1;

    /* Inches per raster unit */

    dd->ipr[0] = 0.01;
    dd->ipr[1] = 0.01;
#if R_GE_version < 4
    dd->asp = jGDasp;
    dd->newDevStruct = 1;
    
    /* Device capabilities */
    dd->canResizePlot = TRUE;
    dd->canChangeFont = TRUE;
    dd->canRotateText = TRUE;
    dd->canResizeText = TRUE;
#endif
    dd->canClip = TRUE;
    dd->canHAdj = 2;
    dd->canChangeGamma = FALSE;

    dd->startps = 10;
    dd->startcol = xd->col;
    dd->startfill = xd->fill;
    dd->startlty = LTY_SOLID;
    dd->startfont = 1;
    dd->startgamma = gamma_fac;

    dd->deviceSpecific = (void *) xd;

    dd->displayListOn = TRUE;

    return(TRUE);
}


/**
 This allocates an newXGDDesc instance  and sets its default values.
 */
newXGDDesc * Rf_allocNewXGDDeviceDesc(double ps)
{
    newXGDDesc *xd;
    /* allocate new device description */
    if (!(xd = (newXGDDesc*)calloc(1, sizeof(newXGDDesc))))
	return FALSE;

    /* From here on, if we need to bail out with "error", */
    /* then we must also free(xd). */

    /*	Font will load at first use.  */

    if (ps < 6 || ps > 24) ps = 12;
    xd->fontface = -1;
    xd->fontsize = -1;
    xd->basefontface = 1;
    xd->basefontsize = ps;

    return(xd);
}


typedef Rboolean (*XGDDeviceDriverRoutine)(NewDevDesc*, char*, 
					    double, double);

static SEXP gcall;

static char *SaveString(SEXP sxp, int offset)
{
    char *s;
    if(!isString(sxp) || length(sxp) <= offset)
	error("invalid string argument");
    s = R_alloc(strlen(CHAR(STRING_ELT(sxp, offset)))+1, sizeof(char));
    strcpy(s, CHAR(STRING_ELT(sxp, offset)));
    return s;
}

static GEDevDesc* 
Rf_addXGDDevice(char *display, double width, double height, char *host, int port)
{
    NewDevDesc *dev = NULL;
    GEDevDesc *dd;
    
    char *devname="xGD";

#ifdef XGD_DEBUG
    printf("Rf_addXGDDevice(\"%s\", %f, %f, \"%s\", %d)\n",display,width,height,host,port);
#endif

    R_CheckDeviceAvailable();
    BEGIN_SUSPEND_INTERRUPTS {

	/* Allocate and initialize the device driver data */
	if (!(dev = (NewDevDesc*)calloc(1, sizeof(NewDevDesc))))
	    return 0;
#if R_GE_version < 4
	/* Do this for early redraw attempts */
	dev->newDevStruct = 1;
	dev->displayList = R_NilValue;
	/* Make sure that this is initialised before a GC can occur.
	 * This (and displayList) get protected during GC
	 */
	dev->savedSnapshot = R_NilValue;
#endif
	/* Took out the GInit because MOST of it is setting up
	 * R base graphics parameters.  
	 * This is supposed to happen via addDevice now.
	 */
	if (!newXGDDeviceDriver((NewDevDesc*)(dev), display, width, height, host, port))
	  {
	    free(dev);
		error("unable to start device %s", devname);
	    return 0;
	  }
	gsetVar(install(".Device"), mkString(devname), R_NilValue);
	dd = GEcreateDevDesc(dev);
	GEaddDevice(dd);
	GEinitDisplayList(dd);
#ifdef XGD_DEBUG
	printf("XGD> devNum=%d, dd=%x\n", devNumber((DevDesc*) dd), dd);
#endif
    } END_SUSPEND_INTERRUPTS;

    return dd;
}

void resizedXGD(NewDevDesc *dd);

void reloadXGD(int *dn) {
  GEDevDesc *gd = GEgetDevice(*dn);
  if (gd) {
    NewDevDesc *dd=gd->dev;
#ifdef XGD_DEBUG
    printf("reloadXGD: dn=%d, dd=%x\n", *dn, dd);
#endif
    if (dd) resizedXGD(dd);
  }
}

void resizedXGD(NewDevDesc *dd) {
  int devNum;
  newXGDDesc *xd = (newXGDDesc *) dd->deviceSpecific;
  setupXGDfunctions(dd);
#ifdef XGD_DEBUG
  printf("dd->size=%x\n", dd->size);
#endif
  dd->size(&(dd->left), &(dd->right), &(dd->bottom), &(dd->top), dd);
  devNum = ndevNumber(dd);
  if (devNum > 0)
    GEplayDisplayList(GEgetDevice(devNum));
}

SEXP newXGD(char **name, double *w, double *h, char **host, int *port) {
  Rf_addXGDDevice(*name, *w, *h, *host, *port);  

  return R_NilValue;
}
